#include "rdp.h"
#include <stdlib.h>

rdp_t* rdp_new (nextcb cb, gpointer data, gint value)
{

  rdp_t* parser;

  parser = g_malloc (sizeof (rdp_t));
  parser->cb = cb;
  parser->data = data;
  parser->start = symbol_new (FALSE, value);

  parser->rules = NULL;

  parser->buffer = g_list_append (NULL, NULL);

  return parser;

}

void rdp_delete (rdp_t* parser)
{

  g_free (parser->start);
  g_free (parser);

}

gpointer leaf_new (gpointer data)
{
  return g_node_new (data);
}

gpointer tree_new (rule_t* rule)
{
  return g_node_new (rule);
}

gpointer tree_add (gpointer tree, gpointer data)
{
  return g_node_append (tree, data);
}

void tree_delete (gpointer tree)
{
  g_node_destroy (tree);
}

symbol_t* buffer_next (rdp_t* parser, gpointer* attrib)
{

  buffer_t* buffer;

  if (parser->buffer->next == NULL)
    {
      buffer = g_malloc (sizeof (buffer_t));
      buffer->symbol = g_malloc (sizeof (symbol_t));
      buffer->symbol->terminal = TRUE;
      buffer->symbol->value = parser->cb (parser->data, &(buffer->attrib));
      g_list_append (parser->buffer, buffer);
    }

  parser->buffer = g_list_next (parser->buffer);
  buffer = (buffer_t*) parser->buffer->data;

  if (attrib)
    *attrib = buffer->attrib;

  return buffer->symbol;
      
}

gboolean rdp_step (rdp_t* parser, symbol_t* symbol, gpointer* attrib)
{

  GList* l;
  GList* buffer;
  gpointer attr;

  buffer = parser->buffer;

  if (symbol != NULL && symbol->terminal)
    {
      symbol_t* s;
      s = buffer_next (parser, &attr);
      if (!symbol_equal (symbol, s))
	{
	  parser->buffer = buffer;
	  return FALSE;
	}
      *attrib = leaf_new (attr);
      return TRUE;
    }

  for (l = g_list_first (parser->rules); l != NULL; l = g_list_next (l))
    {

      rule_t* rule;

      rule = (rule_t*) l->data;

      if (symbol == NULL || symbol_equal (symbol, rule->left))
	{

	  GList* m;

	  *attrib = tree_new (rule);

	  for (m = g_list_first (rule->right); m != NULL; m = g_list_next (m))
	    {

	      symbol_t* s;

	      s = (symbol_t*) m->data;

	      if (!rdp_step (parser, s, &attr))
		{
		  parser->buffer = buffer;
		  break;
		}

	      tree_add (*attrib, attr);

	    }

	  if (m == NULL)
	    return TRUE;
	  else
	    tree_delete (*attrib);

	}

    }

  return FALSE;

}

gpointer rdp_build (rdp_t* parser)
{

  gpointer attrib;

  if (rdp_step (parser, parser->start, &attrib))
    return attrib;

  return NULL;

}
